@@ -27,9 +27,11 @@ int getopt(int argc, const char *argv[], char *optstring);
 # ifndef DBNTWIN32
 #  define DBNTWIN32
 
-// As of Visual Studio .NET 2003, define WIN32_LEAN_AND_MEAN to avoid redefining LPCBYTE in sqlfront.h
-// Unless it was already defined, undefine it after windows.h has been included.  
-// (windows.h includes a bunch of stuff needed by sqlfront.h.  Bleh.)
+/*
+ * As of Visual Studio .NET 2003, define WIN32_LEAN_AND_MEAN to avoid redefining LPCBYTE in sqlfront.h
+ * Unless it was already defined, undefine it after windows.h has been included.
+ * (windows.h includes a bunch of stuff needed by sqlfront.h.  Bleh.)
+ */
 #  ifndef WIN32_LEAN_AND_MEAN
 #   define WIN32_LEAN_AND_MEAN
 #   define WIN32_LEAN_AND_MEAN_DEFINED_HERE
@@ -44,14 +46,12 @@ int getopt(int argc, const char *argv[], char *optstring);
 #  include <sqlfront.h>
 #  include <sqldb.h>
 
-#endif // DBNTWIN32
+#endif /* DBNTWIN32 */
 # include "win32.microsoft/syb2ms.h"
 #endif
 #endif /* MicrosoftsDbLib */
 
-#if HAVE_CONFIG_H
 #include <config.h>
-#endif
 
 #include <stdio.h>
 #include <assert.h>
@@ -91,8 +91,7 @@ int getopt(int argc, const char *argv[], char *optstring);
 #endif
 #endif /* MicrosoftsDbLib */
 
-static char software_version[] = "$Id: defncopy.c,v 1.23.2.1 2011/04/11 09:59:00 freddy77 Exp $";
-static void *no_unused_var_warn[] = { software_version, no_unused_var_warn };
+#include <freetds/sysdep_private.h>
 
 #ifndef MicrosoftsDbLib
 static int err_handler(DBPROCESS * dbproc, int severity, int dberr, int oserr, char *dberrstr, char *oserrstr);
@@ -196,8 +195,11 @@ main(int argc, char *argv[])
 	 * Connect to the server 
 	 */
 	dbproc = dbopen(login, options.servername);
-	assert(dbproc != NULL);
-	
+	if (!dbproc) {
+		fprintf(stderr, "There was a problem connecting to the server.\n");
+		exit(1);
+	}
+
 	/* Switch to the specified database, if any */
 	if (options.database)
 		dbuse(dbproc, options.database);
@@ -212,11 +214,11 @@ main(int argc, char *argv[])
 		static const char query[] = " select	cast(c.text as text)"
 #endif /* MicrosoftsDbLib */
 					 ", number "
-					 " from	syscomments  as c"
-					 " join 	sysobjects as o"
-					 " on 		o.id = c.id"
-					 " where	o.name = '%s'"
-					 " and 	o.uid = user_id('%s')"
+					 " from syscomments c,"
+					 "      sysobjects o"
+					 " where	o.id = c.id"
+					 " and		o.name = '%s'"
+					 " and		o.uid = user_id('%s')"
 					 " and		o.type not in ('U', 'S')" /* no user or system tables */
 					 " order by 	c.number, c.colid"
 					;
@@ -281,10 +283,10 @@ parse_argument(const char argument[], PROCEDURE* procedure)
 		memcpy(procedure->owner, argument, len);
 		procedure->owner[len] = '\0';
 
-		tds_strlcpy(procedure->name, s+1, sizeof(procedure->name));
+		strlcpy(procedure->name, s+1, sizeof(procedure->name));
 	} else {
 		strcpy(procedure->owner, "dbo");
-		tds_strlcpy(procedure->name, argument, sizeof(procedure->name));
+		strlcpy(procedure->name, argument, sizeof(procedure->name));
 	}
 }
 
@@ -327,6 +329,7 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 	RETCODE erc;
 	int row_code, iresultset, i, ret;
 	int maxnamelen = 0, nrows = 0;
+	char **p_str;
 	
 	create_index = tmpfile();
 	
@@ -338,7 +341,7 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 	for (iresultset=1; (erc = dbresults(dbproc)) != NO_MORE_RESULTS; iresultset++) {
 		if (erc == FAIL) {
 			fprintf(stderr, "%s:%d: dbresults(), result set %d failed\n", options.appname, __LINE__, iresultset);
-			return 0;
+			goto cleanup;
 		}
 
 		/* Get the data */
@@ -357,19 +360,19 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 
 				/* name */
 				datlen = dbdatlen(dbproc, 1);
-				index_name = calloc(1, 1 + datlen);
+				index_name = (char *) calloc(1, 1 + datlen);
 				assert(index_name);
 				memcpy(index_name, dbdata(dbproc, 1), datlen);
 				
 				/* kind */
 				datlen = dbdatlen(dbproc, 2);
-				index_description = calloc(1, 1 + datlen);
+				index_description = (char *) calloc(1, 1 + datlen);
 				assert(index_description);
 				memcpy(index_description, dbdata(dbproc, 2), datlen);
 				
 				/* columns */
 				datlen = dbdatlen(dbproc, 3);
-				index_keys = calloc(1, 1 + datlen);
+				index_keys = (char *) calloc(1, 1 + datlen);
 				assert(index_keys);
 				memcpy(index_keys, dbdata(dbproc, 3), datlen);
 				
@@ -422,7 +425,7 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 			colmap = (0 == strcmp("Computed", dbcolname(dbproc, 3)))? microsoft_colmap : sybase_colmap;
 				
 			/* Make room for the next row */
-			p = realloc(ddl, ++nrows * sizeof(struct DDL));
+			p = (struct DDL *) realloc(ddl, ++nrows * sizeof(struct DDL));
 			if (p == NULL) {
 				perror("error: insufficient memory for row DDL");
 				assert(p !=  NULL);
@@ -448,7 +451,7 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 					continue;
 				}
 
-				*coldesc[i] = calloc(1, 1 + datlen); /* calloc will null terminate */
+				*coldesc[i] = (char *) calloc(1, 1 + datlen); /* calloc will null terminate */
 				if( *coldesc[i] == NULL ) {
 					perror("error: insufficient memory for row detail");
 					assert(*coldesc[i] != NULL);
@@ -470,10 +473,9 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 	 * We've collected the description for the columns in the 'ddl' array.  
 	 * Now we'll print the CREATE TABLE statement in jkl's preferred format.  
 	 */
-	if (nrows == 0) {
-		fclose(create_index);
-		return nrows;
-	}
+	if (nrows == 0)
+		goto cleanup;
+
 	printf("%sCREATE TABLE %s.%s\n", use_statement, procedure->owner, procedure->name);
 	for (i=0; i < nrows; i++) {
 		static const char *varytypenames[] =    { "char"  		
@@ -494,6 +496,7 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 		int is_null;
 
 		/* get size of decimal, numeric, char, and image types */
+		ret = 0;
 		if (0 == strcasecmp("decimal", ddl[i].type) || 0 == strcasecmp("numeric", ddl[i].type)) {
 			if (ddl[i].precision && 0 != strcasecmp("NULL", ddl[i].precision)) {
 				rtrim(ddl[i].precision);
@@ -508,6 +511,7 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 				}
 			}
 		}
+		assert(ret >= 0);
 
 		if (colmap == sybase_colmap) 
 			is_null = *(int*)ddl[i].nullable == 1;
@@ -527,7 +531,12 @@ print_ddl(DBPROCESS *dbproc, PROCEDURE *procedure)
 	while ((i = fgetc(create_index)) != EOF) {
 		fputc(i, stdout);
 	}
-	
+
+cleanup:
+	p_str = (char **) ddl;
+	for (i=0; i < nrows * (sizeof(struct DDL)/sizeof(char*)); ++i)
+		free(p_str[i]);
+	free(ddl);
 	fclose(create_index);
 	return nrows;
 }
@@ -543,7 +552,7 @@ print_results(DBPROCESS *dbproc)
 	
 	/* bound variables */
 	enum column_id { ctext=1, number=2 };
-	char sql_text[8000];
+	char sql_text[16000];
 	int	 sql_text_status;
 	int	 procedure_number; /* for create proc abc;2 */
 	int	 procedure_number_status;
@@ -564,11 +573,11 @@ print_results(DBPROCESS *dbproc)
 		/* 
 		 * Bind the columns to our variables.  
 		 */
-		if (sizeof(sql_text) < dbcollen(dbproc, ctext) ) {
-			assert(sizeof(sql_text) >= dbcollen(dbproc, ctext));
+		if (sizeof(sql_text) <= dbcollen(dbproc, ctext) ) {
+			assert(sizeof(sql_text) > dbcollen(dbproc, ctext));
 			return 0;
 		}
-		erc = dbbind(dbproc, ctext, NTBSTRINGBIND, sizeof(sql_text), (BYTE *) sql_text);
+		erc = dbbind(dbproc, ctext, STRINGBIND, 0, (BYTE *) sql_text);
 		if (erc == FAIL) {
 			fprintf(stderr, "%s:%d: dbbind(), column %d failed\n", options.appname, __LINE__, ctext);
 			return -1;
@@ -654,6 +663,7 @@ static LOGINREC *
 get_login(int argc, char *argv[], OPTIONS *options)
 {
 	LOGINREC *login;
+	char *password;
 	int ch;
 	int fdomain = TRUE;
 
@@ -662,7 +672,7 @@ get_login(int argc, char *argv[], OPTIONS *options)
 
 	assert(options && argv);
 	
-	options->appname = tds_basename(argv[0]);
+	options->appname = basename(argv[0]);
 	
 	login = dblogin();
 	
@@ -684,8 +694,11 @@ get_login(int argc, char *argv[], OPTIONS *options)
 			fdomain = FALSE;
 			break;
 		case 'P':
-			DBSETLPWD(login, optarg);
-			memset(optarg, 0, strlen(optarg));
+			password = tds_getpassarg(optarg);
+			DBSETLPWD(login, password);
+			memset(password, 0, strlen(password));
+			free(password);
+			password = NULL;
 			fdomain = FALSE;
 			break;
 		case 'S':
@@ -702,7 +715,7 @@ get_login(int argc, char *argv[], OPTIONS *options)
 			options->output_filename = strdup(optarg);
 			break;
 		case 'v':
-			printf("%s\n\n%s", software_version, 
+			printf("%s\n\n%s", argv[0],
 				"Copyright (C) 2004  James K. Lowden\n"
 				"This program  is free software; you can redistribute it and/or\n"
 				"modify it under the terms of the GNU General Public\n"
