@@ -18,9 +18,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#if HAVE_CONFIG_H
 #include <config.h>
-#endif /* HAVE_CONFIG_H */
 
 #if HAVE_STDLIB_H
 #include <stdlib.h>
@@ -32,16 +30,12 @@
 
 #include <assert.h>
 
-#include "tdsodbc.h"
-#include "tdsiconv.h"
-#include "tdsstring.h"
-#include "tdsconvert.h"
-
-#ifdef DMALLOC
-#include <dmalloc.h>
-#endif
-
-TDS_RCSID(var, "$Id: odbc_util.c,v 1.124 2010/08/04 11:01:34 freddy77 Exp $");
+#include <freetds/odbc.h>
+#include <freetds/iconv.h>
+#include <freetds/string.h>
+#include <freetds/convert.h>
+#include <freetds/enum_cap.h>
+#include <odbcss.h>
 
 /**
  * \ingroup odbc_api
@@ -55,29 +49,14 @@ TDS_RCSID(var, "$Id: odbc_util.c,v 1.124 2010/08/04 11:01:34 freddy77 Exp $");
  */
 
 #ifdef ENABLE_ODBC_WIDE
-static char *odbc_iso2utf(const char *s, int len);
-static char *odbc_mb2utf(TDS_DBC *dbc, const char *s, int len);
-static char *odbc_wide2utf(const SQLWCHAR *s, int len);
+static DSTR *odbc_iso2utf(DSTR *res, const char *s, unsigned int len);
+static DSTR *odbc_mb2utf(TDS_DBC *dbc, DSTR *res, const char *s, unsigned int len);
+static DSTR *odbc_wide2utf(DSTR *res, const SQLWCHAR *s, unsigned int len);
 #endif
 
-static char *
-odbc_strndup(const char *s, int len)
+int
+odbc_set_stmt_query(TDS_STMT * stmt, const ODBC_CHAR *sql, int sql_len _WIDE)
 {
-	char *out = (char*) malloc(len+1);
-	if (!out)
-		return NULL;
-	memcpy(out, s, len);
-	out[len] = 0;
-	return out;
-}
-
-static int
-odbc_set_stmt(TDS_STMT * stmt, char **dest, const ODBC_CHAR *sql, int sql_len _WIDE)
-{
-	char *p;
-
-	assert(dest == &stmt->prepared_query || dest == &stmt->query);
-
 	if (sql_len == SQL_NTS)
 #ifdef ENABLE_ODBC_WIDE
 		sql_len = wide ? sqlwcslen(sql->wide) : strlen(sql->mb);
@@ -92,6 +71,7 @@ odbc_set_stmt(TDS_STMT * stmt, char **dest, const ODBC_CHAR *sql, int sql_len _W
 	stmt->params = NULL;
 	stmt->param_num = 0;
 	stmt->param_count = 0;
+	stmt->is_prepared_query = 0;
 	stmt->prepared_query_is_func = 0;
 	stmt->prepared_query_is_rpc = 0;
 	stmt->prepared_pos = NULL;
@@ -99,38 +79,14 @@ odbc_set_stmt(TDS_STMT * stmt, char **dest, const ODBC_CHAR *sql, int sql_len _W
 	stmt->num_param_rows = 1;
 	stmt->need_reprepare = 0;
 
-	if (stmt->prepared_query)
-		TDS_ZERO_FREE(stmt->prepared_query);
-
-	if (stmt->query)
-		TDS_ZERO_FREE(stmt->query);
-
-#ifdef ENABLE_ODBC_WIDE
-	*dest = p = wide ? odbc_wide2utf(sql->wide, sql_len) : odbc_mb2utf(stmt->dbc, sql->mb, sql_len);
-#else
-	*dest = p = odbc_strndup((const char*) sql, sql_len);
-#endif
-	if (!p)
+	if (!odbc_dstr_copy(stmt->dbc, &stmt->query, sql_len, sql))
 		return SQL_ERROR;
 
 	return SQL_SUCCESS;
 }
 
-int
-odbc_set_stmt_query(TDS_STMT * stmt, const ODBC_CHAR *sql, int sql_len _WIDE)
-{
-	return odbc_set_stmt(stmt, &stmt->query, sql, sql_len _wide);
-}
-
-
-int
-odbc_set_stmt_prepared_query(TDS_STMT * stmt, const ODBC_CHAR *sql, int sql_len _WIDE)
-{
-	return odbc_set_stmt(stmt, &stmt->prepared_query, sql, sql_len _wide);
-}
-
-int
-odbc_get_string_size(int size, ODBC_CHAR * str _WIDE)
+unsigned int
+odbc_get_string_size(int size, const ODBC_CHAR * str _WIDE)
 {
 	if (str) {
 		if (size == SQL_NTS)
@@ -140,26 +96,27 @@ odbc_get_string_size(int size, ODBC_CHAR * str _WIDE)
 			return strlen((const char*) str);
 #endif
 		if (size >= 0)
-			return size;
+			return (unsigned int) size;
 	}
 	/* SQL_NULL_DATA or any other strange value */
 	return 0;
 }
 
 #ifdef ENABLE_ODBC_WIDE
-static char *
-odbc_iso2utf(const char *s, int len)
+static DSTR*
+odbc_iso2utf(DSTR *res, const char *s, unsigned int len)
 {
-	int i, o_len = len + 1;
+	unsigned int i, o_len = len + 1;
 	char *out, *p;
 
-	assert(s && len >= 0);
+	assert(s);
 	for (i = 0; i < len; ++i)
 		if ((s[i] & 0x80) != 0)
 			++o_len;
 
-	out = (char *) malloc(o_len);
-	if (!out) return NULL;
+	if (!tds_dstr_alloc(res, o_len))
+		return NULL;
+	out = tds_dstr_buf(res);
 
 	for (p = out; len > 0; --len) {
 		unsigned char u = (unsigned char) *s++;
@@ -170,23 +127,23 @@ odbc_iso2utf(const char *s, int len)
 			*p++ = u;
 		}
 	}
-	*p = 0;
 	assert(p+1-out <= o_len);
-	return out;
+	return tds_dstr_setlen(res, p - out);
 }
 
-static char *
-odbc_wide2utf(const SQLWCHAR *s, int len)
+static DSTR*
+odbc_wide2utf(DSTR *res, const SQLWCHAR *s, unsigned int len)
 {
-	int i, o_len = len + 1;
+	unsigned int i, o_len = len + 1;
 	char *out, *p;
+
 #if SIZEOF_SQLWCHAR > 2
 # define MASK(n) ((0xffffffffu << (n)) & 0xffffffffu)
 #else
 # define MASK(n) ((0xffffu << (n)) & 0xffffu)
 #endif
 
-	assert(len >= 0 && (s || len == 0));
+	assert(s || len == 0);
 	for (i = 0; i < len; ++i) {
 		if ((s[i] & MASK(7)) == 0)
 			continue;
@@ -207,13 +164,14 @@ odbc_wide2utf(const SQLWCHAR *s, int len)
 #endif
 	}
 
-	out = (char *) malloc(o_len);
-	if (!out) return NULL;
+	if (!tds_dstr_alloc(res, o_len))
+		return NULL;
+	out = tds_dstr_buf(res);
 
 	for (p = out; len > 0; --len) {
 		SQLWCHAR u = *s++;
 		if ((u & MASK(7)) == 0) {
-			*p++ = u;
+			*p++ = (char) u;
 			continue;
 		}
 		if ((u & MASK(11)) == 0) {
@@ -244,13 +202,12 @@ odbc_wide2utf(const SQLWCHAR *s, int len)
 		}
 		*p++ = 0x80 | (0x3f & u);
 	}
-	*p = 0;
 	assert(p+1-out <= o_len);
-	return out;
+	return tds_dstr_setlen(res, p - out);
 }
 
-static char *
-odbc_mb2utf(TDS_DBC *dbc, const char *s, int len)
+static DSTR*
+odbc_mb2utf(TDS_DBC *dbc, DSTR *res, const char *s, unsigned int len)
 {
 	char *buf;
 
@@ -260,19 +217,19 @@ odbc_mb2utf(TDS_DBC *dbc, const char *s, int len)
 	TDSICONV *char_conv = dbc->mb_conv;
 
 	if (!char_conv)
-		return odbc_iso2utf(s, len);
+		return odbc_iso2utf(res, s, len);
 
 	if (char_conv->flags == TDS_ENCODING_MEMCPY)
-		return odbc_strndup(s, len);
+		return tds_dstr_copyn(res, s, len);
 
 	il = len;
 
 	/* allocate needed buffer (+1 is to exclude 0 case) */
-	ol = il * char_conv->server_charset.max_bytes_per_char / char_conv->client_charset.min_bytes_per_char + 1;
+	ol = il * char_conv->to.charset.max_bytes_per_char / char_conv->from.charset.min_bytes_per_char + 1;
 	assert(ol > 0);
-	buf = (char *) malloc(ol);
-	if (!buf)
+	if (!tds_dstr_alloc(res, ol))
 		return NULL;
+	buf = tds_dstr_buf(res);
 
 	ib = s;
 	ob = buf;
@@ -284,31 +241,26 @@ odbc_mb2utf(TDS_DBC *dbc, const char *s, int len)
 		free(buf);
 		return NULL;
 	}
-	*ob = 0;
-	return buf;
+	return tds_dstr_setlen(res, ob - buf);
 }
 #endif
 
 #ifdef ENABLE_ODBC_WIDE
 DSTR*
-odbc_dstr_copy_flag(TDS_DBC *dbc, DSTR *s, int size, ODBC_CHAR * str, int flag)
+odbc_dstr_copy_flag(TDS_DBC *dbc, DSTR *s, int size, const ODBC_CHAR * str, int flag)
 {
 	int wide = flag&1;
-	int len = odbc_get_string_size((flag&0x21) == 0x21 && size >= 0 ? size/SIZEOF_SQLWCHAR : size, str, wide);
-	char *buf;
+	unsigned int len;
 
+	len = odbc_get_string_size((flag&0x21) == 0x21 && size >= 0 ? size/SIZEOF_SQLWCHAR : size, str, wide);
 	if (wide)
-		buf = odbc_wide2utf(str->wide, len);
-	else
-		buf = odbc_mb2utf(dbc, str->mb, len);
-	if (!buf)
-		return NULL;
+		return odbc_wide2utf(s, str->wide, len);
 
-	return tds_dstr_set(s, buf);
+	return odbc_mb2utf(dbc, s, str->mb, len);
 }
 #else
 DSTR*
-odbc_dstr_copy(TDS_DBC *dbc, DSTR *s, int size, ODBC_CHAR * str)
+odbc_dstr_copy(TDS_DBC *dbc, DSTR *s, int size, const ODBC_CHAR * str)
 {
 	return tds_dstr_copyn(s, (const char *) str, odbc_get_string_size(size, str));
 }
@@ -336,14 +288,15 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 	if (len < 0)
 		len = strlen(s);
 
+	if (cbBuffer < 0)
+		cbBuffer = 0;
+
 #ifdef ENABLE_ODBC_WIDE
 	if ((flag & 1) != 0) {
 		/* wide characters */
 		const unsigned char *p = (const unsigned char*) s;
 		SQLWCHAR *dest = (SQLWCHAR*) buffer;
 
-		if (cbBuffer < 0)
-			cbBuffer = 0;
 		if (flag&0x20)
 			cbBuffer /= SIZEOF_SQLWCHAR;
 #ifndef NDEBUG
@@ -384,7 +337,8 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 			result = SQL_SUCCESS_WITH_INFO;
 		}
 		/* terminate buffer */
-		assert(dest == NULL || dest-(SQLWCHAR*) buffer == out_len);
+		assert(dest == NULL || dest-(SQLWCHAR*) buffer == out_len
+		       || (dest-(SQLWCHAR*) buffer <= out_len && cbBuffer <= 1));
 		if (dest && cbBuffer) {
 			*dest++ = 0;
 			assert(dest-(SQLWCHAR*) buffer <= initial_size);
@@ -397,7 +351,6 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 		const unsigned char *p = (const unsigned char*) s;
 		unsigned char *dest = (unsigned char*) buffer;
 
-		assert(cbBuffer >= 0);
 #ifndef NDEBUG
 		initial_size = cbBuffer;
 #endif
@@ -435,7 +388,8 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 			}
 			result = SQL_SUCCESS_WITH_INFO;
 		}
-		assert(dest == NULL || dest-(unsigned char*) buffer == out_len);
+		assert(dest == NULL || dest-(unsigned char*) buffer == out_len
+		       || (dest-(unsigned char*) buffer <= out_len && cbBuffer <= 1));
 		/* terminate buffer */
 		if (dest && cbBuffer) {
 			*dest++ = 0;
@@ -444,6 +398,7 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 		assert(dest == NULL || dest-(unsigned char*) buffer <= initial_size);
 	} else if (dbc->mb_conv->flags == TDS_ENCODING_MEMCPY) {
 		/* to UTF-8 */
+		out_len = len;
 		if (len >= cbBuffer) {
 			len = cbBuffer - 1;
 			result = SQL_SUCCESS_WITH_INFO;
@@ -467,7 +422,7 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 		/* char_conv is only mostly const */
 		memset((TDS_ERRNO_MESSAGE_FLAGS*) &char_conv->suppress, 0, sizeof(char_conv->suppress));
 		char_conv->suppress.e2big = 1;
-		if (tds_iconv(dbc->tds_socket, char_conv, to_client, &ib, &il, &ob, &ol) == (size_t)-1)
+		if (cbBuffer && tds_iconv(dbc->tds_socket, char_conv, to_client, &ib, &il, &ob, &ol) == (size_t)-1 && errno != E2BIG)
 			result = SQL_ERROR;
 		out_len = cbBuffer - ol;
 		while (result != SQL_ERROR && il) {
@@ -477,17 +432,19 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 			char_conv->suppress.e2big = 1;
 			if (tds_iconv(dbc->tds_socket, char_conv, to_client, &ib, &il, &ob, &ol) == (size_t)-1)
 				result = SQL_ERROR;
+			ol = sizeof(discard) - ol;
+			/* if there are still left space copy the partial conversion */
 			if (out_len < cbBuffer) {
-				int max_copy = out_len - cbBuffer;
-				if (max_copy > sizeof(discard) - ol)
-					max_copy = sizeof(discard) - ol;
+				size_t max_copy = cbBuffer - out_len;
+				if (max_copy > ol)
+					max_copy = ol;
 				memcpy(((char *) buffer) + out_len, discard, max_copy);
 			}
-			out_len += sizeof(discard) - ol;
+			out_len += ol;
 		}
 		if (out_len >= cbBuffer && result != SQL_ERROR)
 			result = SQL_SUCCESS_WITH_INFO;
-		if (buffer && cbBuffer >= 0)
+		if (buffer && cbBuffer > 0)
 			((char *) buffer)[cbBuffer-1 < out_len ? cbBuffer-1:out_len] = 0;
 	}
 #else
@@ -517,7 +474,7 @@ odbc_set_string_flag(TDS_DBC *dbc, SQLPOINTER buffer, SQLINTEGER cbBuffer, void
 void
 odbc_set_return_status(struct _hstmt *stmt, unsigned int n_row)
 {
-	TDSSOCKET *tds = stmt->dbc->tds_socket;
+	TDSSOCKET *tds = stmt->tds;
 
 	/* TODO handle different type results (functions) on mssql2k */
 	if (stmt->prepared_query_is_func && tds->has_status) {
@@ -530,7 +487,7 @@ odbc_set_return_status(struct _hstmt *stmt, unsigned int n_row)
 		if (axd->header.sql_desc_count < 1)
 			return;
 		drec = &axd->records[0];
-		data_ptr = drec->sql_desc_data_ptr;
+		data_ptr = (char*) drec->sql_desc_data_ptr;
 
 		if (axd->header.sql_desc_bind_type != SQL_BIND_BY_COLUMN) {
 			len_offset = axd->header.sql_desc_bind_type * n_row;
@@ -544,7 +501,7 @@ odbc_set_return_status(struct _hstmt *stmt, unsigned int n_row)
 #define LEN(ptr) *((SQLLEN*)(((char*)(ptr)) + len_offset))
 
 		len = odbc_tds2sql(stmt, NULL, SYBINT4, (TDS_CHAR *) & tds->ret_status, sizeof(TDS_INT),
-				   drec->sql_desc_concise_type, (void *) data_ptr, drec->sql_desc_octet_length, NULL);
+				   drec->sql_desc_concise_type, (TDS_CHAR *) data_ptr, drec->sql_desc_octet_length, NULL);
 		if (len == SQL_NULL_DATA)
 			return /* SQL_ERROR */ ;
 		if (drec->sql_desc_indicator_ptr)
@@ -558,7 +515,7 @@ odbc_set_return_status(struct _hstmt *stmt, unsigned int n_row)
 void
 odbc_set_return_params(struct _hstmt *stmt, unsigned int n_row)
 {
-	TDSSOCKET *tds = stmt->dbc->tds_socket;
+	TDSSOCKET *tds = stmt->tds;
 	TDSPARAMINFO *info = tds->current_results;
 
 	int i_begin = stmt->prepared_query_is_func ? 1 : 0;
@@ -593,7 +550,7 @@ odbc_set_return_params(struct _hstmt *stmt, unsigned int n_row)
 				break;
 		}
 
-		data_ptr = drec_apd->sql_desc_data_ptr;
+		data_ptr = (char*) drec_apd->sql_desc_data_ptr;
 		if (axd->header.sql_desc_bind_type != SQL_BIND_BY_COLUMN) {
 			len_offset = axd->header.sql_desc_bind_type * n_row;
 			if (axd->header.sql_desc_bind_offset_ptr)
@@ -624,7 +581,7 @@ odbc_set_return_params(struct _hstmt *stmt, unsigned int n_row)
 		 * Or tests are wrong ??
 		 */
 		len = odbc_tds2sql(stmt, colinfo, tds_get_conversion_type(colinfo->on_server.column_type, colinfo->on_server.column_size), src, srclen,
-				   c_type, (void *) data_ptr, drec_apd->sql_desc_octet_length, drec_ipd);
+				   c_type, (TDS_CHAR*) data_ptr, drec_apd->sql_desc_octet_length, drec_ipd);
 		if (len == SQL_NULL_DATA)
 			return /* SQL_ERROR */ ;
 		if (drec_apd->sql_desc_indicator_ptr)
@@ -635,137 +592,12 @@ odbc_set_return_params(struct _hstmt *stmt, unsigned int n_row)
 	}
 }
 
-/**
- * Convert type from database to ODBC
- */
-SQLSMALLINT
-odbc_server_to_sql_type(int col_type, int col_size)
-{
-	/* FIXME finish */
-	switch ((TDS_SERVER_TYPE) col_type) {
-	case XSYBCHAR:
-	case SYBCHAR:
-		return SQL_CHAR;
-	case XSYBVARCHAR:
-	case SYBVARCHAR:
-		return SQL_VARCHAR;
-	case SYBTEXT:
-		return SQL_LONGVARCHAR;
-	case XSYBNCHAR:
-		return SQL_WCHAR;
-	/* TODO really sure ?? SYBNVARCHAR sybase only ?? */
-	case SYBNVARCHAR:
-	case XSYBNVARCHAR:
-		return SQL_WVARCHAR;
-	case SYBNTEXT:
-		return SQL_WLONGVARCHAR;
-	case SYBBIT:
-	case SYBBITN:
-		return SQL_BIT;
-#if (ODBCVER >= 0x0300)
-	case SYB5INT8:
-	case SYBINT8:
-		/* TODO return numeric for odbc2 and convert bigint to numeric */
-		return SQL_BIGINT;
-#endif
-	case SYBINT4:
-		return SQL_INTEGER;
-	case SYBINT2:
-		return SQL_SMALLINT;
-	case SYBINT1:
-		return SQL_TINYINT;
-	case SYBINTN:
-		switch (col_size) {
-		case 1:
-			return SQL_TINYINT;
-		case 2:
-			return SQL_SMALLINT;
-		case 4:
-			return SQL_INTEGER;
-#if (ODBCVER >= 0x0300)
-		case 8:
-			return SQL_BIGINT;
-#endif
-		}
-		break;
-	case SYBREAL:
-		return SQL_REAL;
-	case SYBFLT8:
-		return SQL_DOUBLE;
-	case SYBFLTN:
-		switch (col_size) {
-		case 4:
-			return SQL_REAL;
-		case 8:
-			return SQL_DOUBLE;
-		}
-		break;
-	case SYBMONEY:
-	case SYBMONEY4:
-	case SYBMONEYN:
-		return SQL_DECIMAL;
-	case SYBDATETIME:
-	case SYBDATETIME4:
-	case SYBDATETIMN:
-#if (ODBCVER >= 0x0300)
-		return SQL_TYPE_TIMESTAMP;
-#else
-		return SQL_TIMESTAMP;
-#endif
-	case XSYBBINARY:
-	case SYBBINARY:
-		return SQL_BINARY;
-	case SYBLONGBINARY:
-	case SYBIMAGE:
-		return SQL_LONGVARBINARY;
-	case XSYBVARBINARY:
-	case SYBVARBINARY:
-		return SQL_VARBINARY;
-	case SYBNUMERIC:
-	case SYBDECIMAL:
-		return SQL_NUMERIC;
-#if (ODBCVER >= 0x0300)
-	case SYBUNIQUE:
-#ifdef SQL_GUID
-		return SQL_GUID;
-#else
-		return SQL_CHAR;
-#endif
-	case SYBVARIANT:
-		break;
-#endif
-	case SYBMSXML:
-		return SQL_CHAR;
-		/*
-		 * TODO what should I do with these types ?? 
-		 * return other types can cause additional problems
-		 */
-	case SYBVOID:
-	case SYBSINT1:
-	case SYBUINT2:
-	case SYBUINT4:
-	case SYBUINT8:
-	case SYBUINT1:
-	case SYBDATE:
-	case SYBDATEN:
-	case SYBINTERVAL:
-	case SYBTIME:
-	case SYBTIMEN:
-	case SYBUINTN:
-	case SYBUNITEXT:
-	case SYBXML:
-	case SYBMSUDT:
-		break;
-	}
-	return SQL_UNKNOWN_TYPE;
-}
-
 /**
  * Pass this an SQL_C_* type and get a SYB* type which most closely corresponds
  * to the SQL_C_* type.
- * This function can return XSYBNVARCHAR even if server do not support it
+ * This function can return XSYBNVARCHAR or SYBUINTx even if server do not support it
  */
-int
+TDS_SERVER_TYPE
 odbc_c_to_server_type(int c_type)
 {
 	switch (c_type) {
@@ -786,24 +618,27 @@ odbc_c_to_server_type(int c_type)
 	case SQL_C_BIT:
 		return SYBBIT;
 #if (ODBCVER >= 0x0300)
-	case SQL_C_SBIGINT:
 	case SQL_C_UBIGINT:
+		return SYBUINT8;
+	case SQL_C_SBIGINT:
 		return SYBINT8;
 #ifdef SQL_C_GUID
 	case SQL_C_GUID:
 		return SYBUNIQUE;
 #endif
 #endif
+	case SQL_C_ULONG:
+		return SYBUINT4;
 	case SQL_C_LONG:
 	case SQL_C_SLONG:
-	case SQL_C_ULONG:
 		return SYBINT4;
+	case SQL_C_USHORT:
+		return SYBUINT2;
 	case SQL_C_SHORT:
 	case SQL_C_SSHORT:
-	case SQL_C_USHORT:
 		return SYBINT2;
-	case SQL_C_TINYINT:
 	case SQL_C_STINYINT:
+	case SQL_C_TINYINT:
 	case SQL_C_UTINYINT:
 		return SYBINT1;
 		/* ODBC date formats are completely differect from SQL one */
@@ -813,7 +648,7 @@ odbc_c_to_server_type(int c_type)
 	case SQL_C_TYPE_DATE:
 	case SQL_C_TYPE_TIME:
 	case SQL_C_TYPE_TIMESTAMP:
-		return SYBDATETIME;
+		return SYBMSDATETIME2;
 		/* ODBC numeric/decimal formats are completely differect from tds one */
 	case SQL_C_NUMERIC:
 		return SYBNUMERIC;
@@ -833,198 +668,7 @@ odbc_c_to_server_type(int c_type)
 	case SQL_C_INTERVAL_MINUTE_TO_SECOND:
 		break;
 	}
-	return TDS_FAIL;
-}
-
-void
-odbc_set_sql_type_info(TDSCOLUMN * col, struct _drecord *drec, SQLINTEGER odbc_ver)
-{
-#define SET_INFO(type, prefix, suffix) do { \
-	drec->sql_desc_literal_prefix = prefix; \
-	drec->sql_desc_literal_suffix = suffix; \
-	drec->sql_desc_type_name = type; \
-	return; \
-	} while(0)
-#define SET_INFO2(type, prefix, suffix, len) do { \
-	drec->sql_desc_length = (len); \
-	SET_INFO(type, prefix, suffix); \
-	} while(0)
-
-	drec->sql_desc_octet_length = drec->sql_desc_length = col->on_server.column_size;
-
-	switch (tds_get_conversion_type(col->column_type, col->column_size)) {
-	case XSYBCHAR:
-	case SYBCHAR:
-		if (col->on_server.column_type == XSYBNCHAR)
-			SET_INFO2("nchar", "'", "'", col->on_server.column_size / 2);
-		SET_INFO("char", "'", "'");
-	case XSYBVARCHAR:
-	case SYBVARCHAR:
-		if (col->on_server.column_type == SYBNVARCHAR || col->on_server.column_type == XSYBNVARCHAR)
-			SET_INFO2("nvarchar", "'", "'", col->on_server.column_size / 2);
-		SET_INFO("varchar", "'", "'");
-	case SYBTEXT:
-		if (col->on_server.column_type == SYBNTEXT)
-			SET_INFO2("ntext", "'", "'", col->on_server.column_size / 2);
-		SET_INFO("text", "'", "'");
-	case SYBBIT:
-	case SYBBITN:
-		SET_INFO2("bit", "", "", 1);
-#if (ODBCVER >= 0x0300)
-	case SYBINT8:
-		/* TODO return numeric for odbc2 and convert bigint to numeric */
-		SET_INFO2("bigint", "", "", 19);
-#endif
-	case SYBINT4:
-		SET_INFO2("int", "", "", 10);
-	case SYBINT2:
-		SET_INFO2("smallint", "", "", 5);
-	case SYBINT1:
-		SET_INFO2("tinyint", "", "", 3);
-	case SYBREAL:
-		SET_INFO2("real", "", "", odbc_ver == SQL_OV_ODBC3 ? 24 : 7);
-	case SYBFLT8:
-		SET_INFO2("float", "", "", odbc_ver == SQL_OV_ODBC3 ? 53 : 15);
-	case SYBMONEY:
-		drec->sql_desc_octet_length = 21;
-		SET_INFO2("money", "$", "", 19);
-	case SYBMONEY4:
-		drec->sql_desc_octet_length = 12;
-		SET_INFO2("money", "$", "", 10);
-	case SYBDATETIME:
-		drec->sql_desc_octet_length = sizeof(TIMESTAMP_STRUCT);
-		SET_INFO2("datetime", "'", "'", 23);
-	case SYBDATETIME4:
-		drec->sql_desc_octet_length = sizeof(TIMESTAMP_STRUCT);
-		SET_INFO2("datetime", "'", "'", 16);
-	case SYBBINARY:
-		/* handle TIMESTAMP using usertype */
-		if (col->column_usertype == 80)
-			SET_INFO("timestamp", "0x", "");
-		SET_INFO("binary", "0x", "");
-	case SYBIMAGE:
-		SET_INFO("image", "0x", "");
-	case SYBVARBINARY:
-		SET_INFO("varbinary", "0x", "");
-	case SYBNUMERIC:
-		drec->sql_desc_octet_length = col->column_prec + 2;
-		SET_INFO2("numeric", "", "", col->column_prec);
-	case SYBDECIMAL:
-		drec->sql_desc_octet_length = col->column_prec + 2;
-		SET_INFO2("decimal", "", "", col->column_prec);
-	case SYBINTN:
-	case SYBDATETIMN:
-	case SYBFLTN:
-	case SYBMONEYN:
-		assert(0);
-	case SYBVOID:
-	case SYBNTEXT:
-	case SYBNVARCHAR:
-	case XSYBNVARCHAR:
-	case XSYBNCHAR:
-		break;
-#if (ODBCVER >= 0x0300)
-	case SYBUNIQUE:
-		/* FIXME for Sybase ?? */
-		SET_INFO2("uniqueidentifier", "'", "'", 36);
-	case SYBVARIANT:
-		SET_INFO("sql_variant", "", "");
-		break;
-#endif
-	case SYBMSXML:
-		SET_INFO("xml", "'", "'");
-	}
-	SET_INFO("", "", "");
-#undef SET_INFO
-#undef SET_INFO2
-}
-
-SQLINTEGER
-odbc_sql_to_displaysize(int sqltype, TDSCOLUMN *col)
-{
-	SQLINTEGER size = 0;
-
-	switch (sqltype) {
-	case SQL_CHAR:
-	case SQL_VARCHAR:
-	case SQL_LONGVARCHAR:
-		size = col->on_server.column_size;
-		break;
-	/* FIXME sure ?? *2 or not ?? */
-	case SQL_WCHAR:
-	case SQL_WVARCHAR:
-	case SQL_WLONGVARCHAR:
-		size = col->on_server.column_size / 2;
-		break;
-	case SQL_BINARY:
-	case SQL_VARBINARY:
-	case SQL_LONGVARBINARY:
-		size = col->column_size * 2;
-		break;
-	case SQL_BIGINT:
-		size = 20;
-		break;
-	case SQL_INTEGER:
-		size = 11;	/* -1000000000 */
-		break;
-	case SQL_SMALLINT:
-		size = 6;	/* -10000 */
-		break;
-	case SQL_BIT:
-		size = 1;
-		break;
-	case SQL_TINYINT:
-		size = 3;	/* 255 */
-		break;
-	case SQL_DECIMAL:
-	case SQL_NUMERIC:
-		/* TODO check money format returned by propretary ODBC, scale == 4 but we use 2 digits */
-		if (col->column_type == SYBMONEY || (col->column_type == SYBMONEYN && col->column_size == 8))
-			size = 21;
-		else if (col->column_type == SYBMONEY4 || (col->column_type == SYBMONEYN && col->column_size == 4))
-			size = 12;
-		else
-			size = col->column_prec + 2;
-		break;
-	case SQL_DATE:
-	case SQL_TYPE_DATE:
-		/* FIXME check always yyyy-mm-dd ?? */
-		size = 19;
-		break;
-	case SQL_TIME:
-	case SQL_TYPE_TIME:
-		/* FIXME check always hh:mm:ss[.fff] */
-		size = 19;
-		break;
-	case SQL_TYPE_TIMESTAMP:
-	case SQL_TIMESTAMP:
-		/* TODO dependent on precision (decimal second digits) */
-		/* we always format using yyyy-mm-dd hh:mm:ss[.fff], see convert_tds2sql.c */
-		size = 19;
-		if (col->column_type == SYBDATETIME || (col->column_type == SYBDATETIMN && col->column_size == 8))
-			size = 23;
-		break;
-	case SQL_FLOAT:
-	case SQL_REAL:
-	case SQL_DOUBLE:
-		/* TODO check REAL/FLOAT format */
-		if (col->column_type == SYBREAL || (col->column_type == SYBFLTN && col->column_size == 4))
-			size = 14;
-		else 
-			size = 24;	/* FIXME -- what should the correct size be? */
-		break;
-#ifdef SQL_GUID
-	case SQL_GUID:
-		size = 36;
-		break;
-#endif
-	default:
-		/* FIXME TODO finish, should support ALL types (interval, binary) */
-		size = 40;
-		tdsdump_log(TDS_DBG_INFO1, "odbc_sql_to_displaysize: unknown sql type %d\n", (int) sqltype);
-		break;
-	}
-	return size;
+	return TDS_INVALID_TYPE;
 }
 
 int
@@ -1086,24 +730,34 @@ odbc_sql_to_c_type_default(int sql_type)
 	}
 }
 
-int
-odbc_sql_to_server_type(TDSSOCKET * tds, int sql_type)
+TDS_SERVER_TYPE
+odbc_sql_to_server_type(TDSCONNECTION * conn, int sql_type, int sql_unsigned)
 {
 
 	switch (sql_type) {
 	case SQL_WCHAR:
-		if (IS_TDS7_PLUS(tds))
+		if (IS_TDS7_PLUS(conn))
 			return XSYBNCHAR;
 	case SQL_CHAR:
 		return SYBCHAR;
 	case SQL_WVARCHAR:
-		if (IS_TDS7_PLUS(tds))
+		if (IS_TDS7_PLUS(conn))
 			return XSYBNVARCHAR;
 	case SQL_VARCHAR:
 		return SYBVARCHAR;
+	case SQL_SS_VARIANT:
+		if (IS_TDS71_PLUS(conn))
+			return SYBVARIANT;
+		if (IS_TDS7_PLUS(conn))
+			return XSYBNVARCHAR;
+		return SYBVARCHAR;
+	case SQL_SS_XML:
+		if (IS_TDS72_PLUS(conn))
+			return SYBMSXML;
 	case SQL_WLONGVARCHAR:
-		if (IS_TDS7_PLUS(tds))
+		if (IS_TDS7_PLUS(conn))
 			return SYBNTEXT;
+		/* fall thought */
 	case SQL_LONGVARCHAR:
 		return SYBTEXT;
 	case SQL_DECIMAL:
@@ -1112,21 +766,27 @@ odbc_sql_to_server_type(TDSSOCKET * tds, int sql_type)
 		return SYBNUMERIC;
 #ifdef SQL_GUID
 	case SQL_GUID:
-		if (IS_TDS7_PLUS(tds))
+		if (IS_TDS7_PLUS(conn))
 			return SYBUNIQUE;
-		return 0;
+		return TDS_INVALID_TYPE;
 #endif
 	case SQL_BIT:
-		if (IS_TDS7_PLUS(tds))
+		if (IS_TDS7_PLUS(conn))
 			return SYBBITN;
 		return SYBBIT;
 	case SQL_TINYINT:
 		return SYBINT1;
 	case SQL_SMALLINT:
+		if (sql_unsigned && tds_capability_has_req(conn, TDS_REQ_DATA_UINT2))
+			return SYBUINT2;
 		return SYBINT2;
 	case SQL_INTEGER:
+		if (sql_unsigned && tds_capability_has_req(conn, TDS_REQ_DATA_UINT4))
+			return SYBUINT4;
 		return SYBINT4;
 	case SQL_BIGINT:
+		if (sql_unsigned && tds_capability_has_req(conn, TDS_REQ_DATA_UINT8))
+			return SYBUINT8;
 		return SYBINT8;
 	case SQL_REAL:
 		return SYBREAL;
@@ -1139,8 +799,38 @@ odbc_sql_to_server_type(TDSSOCKET * tds, int sql_type)
 	case SQL_TIMESTAMP:
 		/* ODBC version 3 */
 	case SQL_TYPE_DATE:
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_BIGDATETIME))
+			return SYB5BIGDATETIME;
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_DATE))
+			return SYBDATE;
+		if (IS_TDS73_PLUS(conn))
+			return SYBMSDATE;
+		goto type_timestamp;
 	case SQL_TYPE_TIME:
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_BIGTIME))
+			return SYB5BIGTIME;
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_TIME))
+			return SYBTIME;
+		if (IS_TDS73_PLUS(conn))
+			return SYBMSTIME;
+	type_timestamp:
 	case SQL_TYPE_TIMESTAMP:
+		if (IS_TDS73_PLUS(conn))
+			return SYBMSDATETIME2;
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_BIGDATETIME))
+			return SYB5BIGDATETIME;
+		return SYBDATETIME;
+	case SQL_SS_TIME2:
+		if (IS_TDS73_PLUS(conn))
+			return SYBMSTIME;
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_BIGDATETIME))
+			return SYB5BIGDATETIME;
+		return SYBDATETIME;
+	case SQL_SS_TIMESTAMPOFFSET:
+		if (IS_TDS73_PLUS(conn))
+			return SYBMSDATETIMEOFFSET;
+		if (IS_TDS50(conn) && tds_capability_has_req(conn, TDS_REQ_DATA_BIGDATETIME))
+			return SYB5BIGDATETIME;
 		return SYBDATETIME;
 	case SQL_BINARY:
 		return SYBBINARY;
@@ -1150,7 +840,7 @@ odbc_sql_to_server_type(TDSSOCKET * tds, int sql_type)
 		return SYBIMAGE;
 		/* TODO interval types */
 	default:
-		return 0;
+		return TDS_INVALID_TYPE;
 	}
 }
 
@@ -1158,8 +848,9 @@ odbc_sql_to_server_type(TDSSOCKET * tds, int sql_type)
 void
 odbc_rdbms_version(TDSSOCKET * tds, char *pversion_string)
 {
-	sprintf(pversion_string, "%.02d.%.02d.%.04d", (int) ((tds->product_version & 0x7F000000) >> 24),
-		(int) ((tds->product_version & 0x00FF0000) >> 16), (int) (tds->product_version & 0x0000FFFF));
+	TDS_UINT version = tds->conn->product_version;
+	sprintf(pversion_string, "%.02d.%.02d.%.04d", (int) ((version & 0x7F000000) >> 24),
+		(int) ((version & 0x00FF0000) >> 16), (int) (version & 0x0000FFFF));
 }
 
 /** Return length of parameter from parameter information */
@@ -1192,13 +883,14 @@ odbc_get_param_len(const struct _drecord *drec_axd, const struct _drecord *drec_
 			len = SQL_NTS;
 		} else {
 			int type = drec_axd->sql_desc_concise_type;
+			TDS_SERVER_TYPE server_type;
 
 			if (type == SQL_C_DEFAULT)
 				type = odbc_sql_to_c_type_default(drec_ixd->sql_desc_concise_type);
-			type = odbc_c_to_server_type(type);
+			server_type = odbc_c_to_server_type(type);
 
 			/* FIXME check what happen to DATE/TIME types */
-			size = tds_get_size_by_type(type);
+			size = tds_get_size_by_type(server_type);
 			if (size > 0)
 				len = size;
 		}
@@ -1238,6 +930,12 @@ odbc_get_param_len(const struct _drecord *drec_axd, const struct _drecord *drec_
 	TYPE_NORMAL(SQL_FLOAT) \
 	TYPE_NORMAL(SQL_REAL) \
 	TYPE_NORMAL(SQL_DOUBLE)\
+\
+	TYPE_NORMAL(SQL_SS_TIMESTAMPOFFSET) \
+	TYPE_NORMAL(SQL_SS_TIME2) \
+	TYPE_NORMAL(SQL_SS_XML) \
+	TYPE_NORMAL(SQL_SS_VARIANT) \
+	TYPE_NORMAL(SQL_TYPE_DATE) \
 \
 	TYPE_VERBOSE_START(SQL_DATETIME) \
 	TYPE_VERBOSE_DATE(SQL_DATETIME, SQL_CODE_TIMESTAMP, SQL_TYPE_TIMESTAMP, SQL_TIMESTAMP) \
@@ -1418,6 +1116,7 @@ odbc_set_concise_c_type(SQLSMALLINT concise_type, struct _drecord * drec, int ch
 
 		switch (drec->sql_desc_type) {
 		case SQL_C_NUMERIC:
+			drec->sql_desc_length = 38;
 			drec->sql_desc_precision = 38;
 			drec->sql_desc_scale = 0;
 			break;
