@@ -1,6 +1,6 @@
 /*
- * This file produced from ../../../src/tds/encodings.pl on Wed Aug 17 22:22:08 2011
- * $Id: encodings.pl,v 1.12 2010/11/26 19:46:55 freddy77 Exp $
+ * This file produced from ./encodings.pl on Tue Jun 28 09:01:13 2016
+ * $Id: encodings.pl,v 1.12 2010-11-26 19:46:55 freddy77 Exp $
  */
 #ifdef TDS_ICONV_ENCODING_TABLES
 
@@ -108,7 +108,6 @@ static const TDS_ENCODING canonic_charsets[] = {
 	{          "UTF-32LE",	4, 4, 100},	/* 100 */
 	{             "UTF-7",	1, 4, 101},	/* 101 */
 	{            "VISCII",	1, 1, 102},	/* 102 */
-	{"",	0, 0, 0}
 };
 
 static const CHARACTER_SET_ALIAS iconv_aliases[] = {
